module DiffFusion

using ChainRulesCore
using DelimitedFiles
using Distributions
using FiniteDifferences
using ForwardDiff
using Interpolations
using LinearAlgebra
using LsqFit
using OrderedCollections
using Printf
using ProgressBars
using QuadGK
using Random
using Roots
using Sobol
using SparseArrays
using Zygote

import Base.length
import Base.string

"""
A type alias for variables representing time.
"""
ModelTime = Number

"""
A type alias for variables representing modelled quantities.
"""
ModelValue = Number

include("termstructures/Termstructures.jl")
include("termstructures/correlation/CorrelationHolder.jl")
include("termstructures/credit/CreditDefaultTermstructures.jl")
include("termstructures/parameter/ParameterTermstructures.jl")
include("termstructures/rates/YieldTermstructures.jl")
include("termstructures/volatility/VolatilityTermstructures.jl")

include("models/Model.jl")
include("models/asset/AssetModel.jl")
include("models/asset/LognormalAssetModel.jl")
include("models/hybrid/CompositeModel.jl")
include("models/hybrid/SimpleModel.jl")
include("models/rates/SeparableHjmModel.jl")
include("models/rates/GaussianHjmModel.jl")
include("models/rates/ForwardRateVolatility.jl")
include("models/rates/SwapRateVolatility.jl")
include("models/rates/SwapRateCalibration.jl")
include("models/futures/MarkovFutureModels.jl")

include("simulations/RandomNumbers.jl")
include("simulations/Simulation.jl")

include("paths/AbstractPath.jl")
include("paths/Context.jl")
include("paths/Path.jl")
include("paths/PathMethods.jl")

include("payoffs/Payoff.jl")
include("payoffs/Leafs.jl")
include("payoffs/UnaryNodes.jl")
include("payoffs/BinaryNodes.jl")
include("payoffs/RatesPayoffs.jl")
include("payoffs/RatesOptions.jl")
include("payoffs/AmcPayoffs.jl")

include("products/Cashflows.jl")
include("products/RatesCoupons.jl")
include("products/CashFlowLeg.jl")
include("products/SwaptionLeg.jl")
include("products/MtMCashFlowLeg.jl")
include("products/CashAndAssetLegs.jl")
include("products/BermudanSwaptionLeg.jl")

include("utils/Bachelier.jl")
include("utils/Black.jl")
include("utils/Gradients.jl")
include("utils/Integrations.jl")
include("utils/InterpolationMethods.jl")
include("utils/PolynomialRegression.jl")
include("utils/PiecewiseRegression.jl")

include("analytics/Scenarios.jl")
include("analytics/Analytics.jl")
include("analytics/Collateral.jl")
include("analytics/Valuations.jl")

include("serialisation/Serialisations.jl")
include("serialisation/Array.jl")
include("serialisation/Termstructures.jl")
include("serialisation/Models.jl")
include("serialisation/RebuildModels.jl")
include("serialisation/RebuildTermstructures.jl")
include("serialisation/ReadTermstructures.jl")

module Examples
    using DiffFusion
    using DiffFusion:ModelTime
    using DiffFusion:ModelValue
    using OrderedCollections
    using Random
    using YAML
    #
    include("examples/csv/csv.jl")
    include("examples/yaml/yaml.jl")
    include("examples/Examples.jl")
    include("examples/Models.jl")
    include("examples/Products.jl")
end # module

include("chainrules/models.jl")
include("chainrules/simulations.jl")

"List of function names eligible for de-serialisation."
const _eligible_func_names = [ string(n) for n in names(DiffFusion; all = true, imported = false) ]

end
